/*!
 * Form tracking for Snowplow v4.6.5 (http://bit.ly/sp-js)
 * Copyright 2022 Snowplow Analytics Ltd, 2010 Anthon Pang
 * Licensed under BSD-3-Clause
 */

(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.snowplowFormTracking = {}));
})(this, (function (exports) { 'use strict';

  /*!
   * Core functionality for Snowplow JavaScript trackers v4.6.5 (http://bit.ly/sp-js)
   * Copyright 2022 Snowplow Analytics Ltd, 2010 Anthon Pang
   * Licensed under BSD-3-Clause
   */

  /*
   * Copyright (c) 2022 Snowplow Analytics Ltd, 2010 Anthon Pang
   * All rights reserved.
   *
   * Redistribution and use in source and binary forms, with or without
   * modification, are permitted provided that the following conditions are met:
   *
   * 1. Redistributions of source code must retain the above copyright notice, this
   *    list of conditions and the following disclaimer.
   *
   * 2. Redistributions in binary form must reproduce the above copyright notice,
   *    this list of conditions and the following disclaimer in the documentation
   *    and/or other materials provided with the distribution.
   *
   * 3. Neither the name of the copyright holder nor the names of its
   *    contributors may be used to endorse or promote products derived from
   *    this software without specific prior written permission.
   *
   * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
   * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
   * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
   * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
   * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
   * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
   * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
   */
  function payloadBuilder() {
      var dict = {}, allJson = [], jsonForProcessing = [], contextEntitiesForProcessing = [];
      var processor;
      var add = function (key, value) {
          if (value != null && value !== '') {
              // null also checks undefined
              dict[key] = value;
          }
      };
      var addDict = function (dict) {
          for (var key in dict) {
              if (Object.prototype.hasOwnProperty.call(dict, key)) {
                  add(key, dict[key]);
              }
          }
      };
      var addJson = function (keyIfEncoded, keyIfNotEncoded, json) {
          if (json && isNonEmptyJson(json)) {
              var jsonWithKeys = { keyIfEncoded: keyIfEncoded, keyIfNotEncoded: keyIfNotEncoded, json: json };
              jsonForProcessing.push(jsonWithKeys);
              allJson.push(jsonWithKeys);
          }
      };
      var addContextEntity = function (entity) {
          contextEntitiesForProcessing.push(entity);
      };
      return {
          add: add,
          addDict: addDict,
          addJson: addJson,
          addContextEntity: addContextEntity,
          getPayload: function () { return dict; },
          getJson: function () { return allJson; },
          withJsonProcessor: function (jsonProcessor) {
              processor = jsonProcessor;
          },
          build: function () {
              processor === null || processor === void 0 ? void 0 : processor(this, jsonForProcessing, contextEntitiesForProcessing);
              return dict;
          },
      };
  }
  /**
   * Is property a non-empty JSON?
   * @param property - Checks if object is non-empty json
   */
  function isNonEmptyJson(property) {
      if (!isJson(property)) {
          return false;
      }
      for (var key in property) {
          if (Object.prototype.hasOwnProperty.call(property, key)) {
              return true;
          }
      }
      return false;
  }
  /**
   * Is property a JSON?
   * @param property - Checks if object is json
   */
  function isJson(property) {
      return (typeof property !== 'undefined' &&
          property !== null &&
          (property.constructor === {}.constructor || property.constructor === [].constructor));
  }
  var LOG_LEVEL;
  (function (LOG_LEVEL) {
      LOG_LEVEL[LOG_LEVEL["none"] = 0] = "none";
      LOG_LEVEL[LOG_LEVEL["error"] = 1] = "error";
      LOG_LEVEL[LOG_LEVEL["warn"] = 2] = "warn";
      LOG_LEVEL[LOG_LEVEL["debug"] = 3] = "debug";
      LOG_LEVEL[LOG_LEVEL["info"] = 4] = "info";
  })(LOG_LEVEL || (LOG_LEVEL = {}));
  /**
   * Find dynamic context generating functions and return their results to be merged into the static contexts
   * Combine an array of unchanging contexts with the result of a context-creating function
   *
   * @param dynamicOrStaticContexts - Array of custom context Objects or custom context generating functions
   * @param Parameters - to pass to dynamic context callbacks
   * @returns An array of Self Describing JSON context
   */
  function resolveDynamicContext(dynamicOrStaticContexts) {
      var _a;
      var extraParams = [];
      for (var _i = 1; _i < arguments.length; _i++) {
          extraParams[_i - 1] = arguments[_i];
      }
      return ((_a = dynamicOrStaticContexts === null || dynamicOrStaticContexts === void 0 ? void 0 : dynamicOrStaticContexts.map(function (context) {
          if (typeof context === 'function') {
              try {
                  return context.apply(void 0, extraParams);
              }
              catch (e) {
                  //TODO: provide warning
                  return undefined;
              }
          }
          else {
              return context;
          }
      }).filter(Boolean)) !== null && _a !== void 0 ? _a : []);
  }
  /**
   * Build a self-describing event
   * A custom event type, allowing for an event to be tracked using your own custom schema
   * and a data object which conforms to the supplied schema
   *
   * @param event - Contains the properties and schema location for the event
   * @returns PayloadBuilder to be sent to {@link @snowplow/tracker-core#TrackerCore.track}
   */
  function buildSelfDescribingEvent(event) {
      var _a = event.event, schema = _a.schema, data = _a.data, pb = payloadBuilder();
      var ueJson = {
          schema: 'iglu:com.snowplowanalytics.snowplow/unstruct_event/jsonschema/1-0-0',
          data: { schema: schema, data: data },
      };
      pb.add('e', 'ue');
      pb.addJson('ue_px', 'ue_pr', ueJson);
      return pb;
  }
  /**
   * Build a Form Focus or Change Form Event based on schema property
   * When a user focuses on a form element or when a user makes a
   * change to a form element.
   *
   * @param event - Contains the properties for the Form Focus or Change Form event
   * @returns PayloadBuilder to be sent to {@link @snowplow/tracker-core#TrackerCore.track}
   */
  function buildFormFocusOrChange(event) {
      var event_schema = '';
      var schema = event.schema, formId = event.formId, elementId = event.elementId, nodeName = event.nodeName, elementClasses = event.elementClasses, value = event.value, type = event.type;
      var event_data = { formId: formId, elementId: elementId, nodeName: nodeName, elementClasses: elementClasses, value: value };
      if (schema === 'change_form') {
          event_schema = 'iglu:com.snowplowanalytics.snowplow/change_form/jsonschema/1-0-0';
          event_data.type = type;
      }
      else if (schema === 'focus_form') {
          event_schema = 'iglu:com.snowplowanalytics.snowplow/focus_form/jsonschema/1-0-0';
          event_data.elementType = type;
      }
      return buildSelfDescribingEvent({
          event: {
              schema: event_schema,
              data: removeEmptyProperties(event_data, { value: true }),
          },
      });
  }
  /**
   * Build a Form Submission Event
   * Used to track when a user submits a form
   *
   * @param event - Contains the properties for the Form Submission event
   * @returns PayloadBuilder to be sent to {@link @snowplow/tracker-core#TrackerCore.track}
   */
  function buildFormSubmission(event) {
      var formId = event.formId, formClasses = event.formClasses, elements = event.elements;
      return buildSelfDescribingEvent({
          event: {
              schema: 'iglu:com.snowplowanalytics.snowplow/submit_form/jsonschema/1-0-0',
              data: removeEmptyProperties({ formId: formId, formClasses: formClasses, elements: elements }),
          },
      });
  }
  /**
   * Returns a copy of a JSON with undefined and null properties removed
   *
   * @param event - JSON object to clean
   * @param exemptFields - Set of fields which should not be removed even if empty
   * @returns A cleaned copy of eventJson
   */
  function removeEmptyProperties(event, exemptFields) {
      if (exemptFields === void 0) { exemptFields = {}; }
      var ret = {};
      for (var k in event) {
          if (exemptFields[k] || (event[k] !== null && typeof event[k] !== 'undefined')) {
              ret[k] = event[k];
          }
      }
      return ret;
  }

  /*!
   * Core functionality for Snowplow Browser trackers v4.6.5 (http://bit.ly/sp-js)
   * Copyright 2022 Snowplow Analytics Ltd, 2010 Anthon Pang
   * Licensed under BSD-3-Clause
   */
  /**
   * Cross-browser helper function to add event handler
   */
  function addEventListener(element, eventType, eventHandler, options) {
      if (element.addEventListener) {
          element.addEventListener(eventType, eventHandler, options);
          return true;
      }
      // IE Support
      if (element.attachEvent) {
          return element.attachEvent('on' + eventType, eventHandler);
      }
      element['on' + eventType] = eventHandler;
  }
  /**
   * Deletes an arbitrary cookie by setting the expiration date to the past
   *
   * @param cookieName - The name of the cookie to delete
   * @param domainName - The domain the cookie is in
   */
  function deleteCookie(cookieName, path, domainName, sameSite, secure) {
      cookie(cookieName, '', -1, path, domainName, sameSite, secure);
  }
  /**
   * Get and set the cookies associated with the current document in browser
   * This implementation always returns a string, returns the cookie value if only name is specified
   *
   * @param name - The cookie name (required)
   * @param value - The cookie value
   * @param ttl - The cookie Time To Live (seconds)
   * @param path - The cookies path
   * @param domain - The cookies domain
   * @param samesite - The cookies samesite attribute
   * @param secure - Boolean to specify if cookie should be secure
   * @returns string The cookies value
   */
  function cookie(name, value, ttl, path, domain, samesite, secure) {
      if (arguments.length > 1) {
          return (document.cookie =
              name +
                  '=' +
                  encodeURIComponent(value !== null && value !== void 0 ? value : '') +
                  (ttl ? '; Expires=' + new Date(+new Date() + ttl * 1000).toUTCString() : '') +
                  (path ? '; Path=' + path : '') +
                  (domain ? '; Domain=' + domain : '') +
                  (samesite ? '; SameSite=' + samesite : '') +
                  (secure ? '; Secure' : ''));
      }
      return decodeURIComponent((('; ' + document.cookie).split('; ' + name + '=')[1] || '').split(';')[0]);
  }
  /**
   * Convert a criterion object to a filter function
   *
   * @param object - criterion Either {allowlist: [array of allowable strings]}
   *                             or {denylist: [array of allowable strings]}
   *                             or {filter: function (elt) {return whether to track the element}
   * @param boolean - byClass Whether to allowlist/denylist based on an element's classes (for forms)
   *                        or name attribute (for fields)
   */
  function getFilterByClass(criterion) {
      // If the criterion argument is not an object, add listeners to all elements
      if (criterion == null || typeof criterion !== 'object' || Array.isArray(criterion)) {
          return function () {
              return true;
          };
      }
      var inclusive = Object.prototype.hasOwnProperty.call(criterion, 'allowlist');
      var specifiedClassesSet = getSpecifiedClassesSet(criterion);
      return getFilter(criterion, function (elt) {
          return checkClass(elt, specifiedClassesSet) === inclusive;
      });
  }
  /**
   * Convert a criterion object to a filter function
   *
   * @param object - criterion Either {allowlist: [array of allowable strings]}
   *                             or {denylist: [array of allowable strings]}
   *                             or {filter: function (elt) {return whether to track the element}
   */
  function getFilterByName(criterion) {
      // If the criterion argument is not an object, add listeners to all elements
      if (criterion == null || typeof criterion !== 'object' || Array.isArray(criterion)) {
          return function () {
              return true;
          };
      }
      var inclusive = criterion.hasOwnProperty('allowlist');
      var specifiedClassesSet = getSpecifiedClassesSet(criterion);
      return getFilter(criterion, function (elt) {
          return elt.name in specifiedClassesSet === inclusive;
      });
  }
  /**
   * List the classes of a DOM element without using elt.classList (for compatibility with IE 9)
   */
  function getCssClasses(elt) {
      return elt.className.match(/\S+/g) || [];
  }
  /**
   * Check whether an element has at least one class from a given list
   */
  function checkClass(elt, classList) {
      var classes = getCssClasses(elt);
      for (var _i = 0, classes_1 = classes; _i < classes_1.length; _i++) {
          var className = classes_1[_i];
          if (classList[className]) {
              return true;
          }
      }
      return false;
  }
  function getFilter(criterion, fallbackFilter) {
      if (criterion.hasOwnProperty('filter') && criterion.filter) {
          return criterion.filter;
      }
      return fallbackFilter;
  }
  function getSpecifiedClassesSet(criterion) {
      // Convert the array of classes to an object of the form {class1: true, class2: true, ...}
      var specifiedClassesSet = {};
      var specifiedClasses = criterion.allowlist || criterion.denylist;
      if (specifiedClasses) {
          if (!Array.isArray(specifiedClasses)) {
              specifiedClasses = [specifiedClasses];
          }
          for (var i = 0; i < specifiedClasses.length; i++) {
              specifiedClassesSet[specifiedClasses[i]] = true;
          }
      }
      return specifiedClassesSet;
  }

  function newCookie(name) {
      var flushTimer;
      var lastSetValueArgs;
      var cacheExpireAt;
      var flushed = true;
      var flushTimeout = 10; // milliseconds
      var maxCacheTtl = 0.05; // seconds
      function getValue() {
          var _a;
          // Note: we can't cache the cookie value as we don't know the expiration date
          if (lastSetValueArgs && (!cacheExpireAt || cacheExpireAt > new Date())) {
              return (_a = lastSetValueArgs[0]) !== null && _a !== void 0 ? _a : cookie(name);
          }
          return cookie(name);
      }
      function setValue(value, ttl, path, domain, samesite, secure) {
          lastSetValueArgs = [value, ttl, path, domain, samesite, secure];
          flushed = false;
          // throttle setting the cookie
          if (flushTimer === undefined) {
              flushTimer = setTimeout(function () {
                  flushTimer = undefined;
                  flush();
              }, flushTimeout);
          }
          cacheExpireAt = new Date(Date.now() + Math.min(maxCacheTtl, ttl !== null && ttl !== void 0 ? ttl : maxCacheTtl) * 1000);
          return true;
      }
      function deleteValue(path, domainName, sameSite, secure) {
          lastSetValueArgs = undefined;
          flushed = true;
          // cancel setting the cookie
          if (flushTimer !== undefined) {
              clearTimeout(flushTimer);
              flushTimer = undefined;
          }
          deleteCookie(name, path, domainName, sameSite, secure);
      }
      function flush() {
          if (flushTimer !== undefined) {
              clearTimeout(flushTimer);
              flushTimer = undefined;
          }
          if (flushed) {
              return;
          }
          flushed = true;
          if (lastSetValueArgs !== undefined) {
              var value = lastSetValueArgs[0], ttl = lastSetValueArgs[1], path = lastSetValueArgs[2], domain = lastSetValueArgs[3], samesite = lastSetValueArgs[4], secure = lastSetValueArgs[5];
              cookie(name, value, ttl, path, domain, samesite, secure);
          }
      }
      return {
          getValue: getValue,
          setValue: setValue,
          deleteValue: deleteValue,
          flush: flush,
      };
  }
  /**
   * Create a new async cookie storage
   *
   * @returns A new cookie storage
   */
  function newCookieStorage() {
      var cache = {};
      function getOrInitCookie(name) {
          if (!cache[name]) {
              cache[name] = newCookie(name);
          }
          return cache[name];
      }
      function getCookie(name) {
          return getOrInitCookie(name).getValue();
      }
      function setCookie(name, value, ttl, path, domain, samesite, secure) {
          return getOrInitCookie(name).setValue(value, ttl, path, domain, samesite, secure);
      }
      function deleteCookie(name, path, domainName, sameSite, secure) {
          getOrInitCookie(name).deleteValue(path, domainName, sameSite, secure);
      }
      function clearCache() {
          cache = {};
      }
      function flush() {
          for (var _i = 0, _a = Object.values(cache); _i < _a.length; _i++) {
              var cookie_1 = _a[_i];
              cookie_1.flush();
          }
      }
      return {
          getCookie: getCookie,
          setCookie: setCookie,
          deleteCookie: deleteCookie,
          clearCache: clearCache,
          flush: flush,
      };
  }
  /**
   * Cookie storage instance with asynchronous cookie writes
   */
  var asyncCookieStorage = newCookieStorage();
  /**
   * Write all pending cookies to the browser.
   * Useful if you track events just before the page is unloaded.
   * This call is not necessary if `synchronousCookieWrite` is set to `true`.
   */
  function flushPendingCookies() {
      asyncCookieStorage.flush();
  }

  /** Events to capture in form tracking */
  var FormTrackingEvent;
  (function (FormTrackingEvent) {
      /** Form field changed event */
      FormTrackingEvent["CHANGE_FORM"] = "change_form";
      /** Form field focused event */
      FormTrackingEvent["FOCUS_FORM"] = "focus_form";
      /** Form submitted event */
      FormTrackingEvent["SUBMIT_FORM"] = "submit_form";
  })(FormTrackingEvent || (FormTrackingEvent = {}));
  var defaultFormTrackingEvents = [
      FormTrackingEvent.CHANGE_FORM,
      FormTrackingEvent.FOCUS_FORM,
      FormTrackingEvent.SUBMIT_FORM,
  ];
  var defaultTransformFn = function (x) { return x; };
  var _focusListeners = {};
  var _changeListeners = {};
  var _submitListeners = {};
  var _targets = {};
  /**
   * Add submission/focus/change event listeners to page for forms and elements according to `configuration`
   *
   * @param tracker The tracker instance the listener belongs to that will be used to track events
   * @param configuration Plugin configuration controlling the events to track and forms/fields to target or transform
   */
  function addFormListeners(tracker, configuration) {
      var _a;
      var options = configuration.options, context = configuration.context, config = getConfigurationForOptions(options);
      var events = (_a = options === null || options === void 0 ? void 0 : options.events) !== null && _a !== void 0 ? _a : defaultFormTrackingEvents;
      var targets = (_targets[tracker.id] = getTargetList(options === null || options === void 0 ? void 0 : options.targets, config.forms));
      if (events.indexOf(FormTrackingEvent.FOCUS_FORM) !== -1) {
          _focusListeners[tracker.id] = getFormChangeListener(tracker, config, FormTrackingEvent.FOCUS_FORM, context);
          targets.forEach(function (target) { return addEventListener(target, 'focus', _focusListeners[tracker.id], true); });
      }
      if (events.indexOf(FormTrackingEvent.CHANGE_FORM) !== -1) {
          _changeListeners[tracker.id] = getFormChangeListener(tracker, config, FormTrackingEvent.CHANGE_FORM, context);
          targets.forEach(function (target) { return addEventListener(target, 'change', _changeListeners[tracker.id], true); });
      }
      if (events.indexOf(FormTrackingEvent.SUBMIT_FORM) !== -1) {
          _submitListeners[tracker.id] = getFormSubmissionListener(tracker, config, context);
          targets.forEach(function (target) { return addEventListener(target, 'submit', _submitListeners[tracker.id], true); });
      }
  }
  /**
   * Builds a list of targets for the plugin event listeners
   *
   * The list can include any specifically provided targets, and will be extended to include the root nodes of any explicit HTMLFormElements provided
   * With neither provided, defaults to the current page's `document` element
   *
   * @param configTargets Explicitly configured list of event target listeners, if any
   * @param forms Explicitly configured list of form elements to track, if any
   * @returns List of EventTargets to add the listener to
   */
  function getTargetList(configTargets, forms) {
      // we attach to document rather than window because the window focus event occurs more often than we require
      var targets = configTargets !== null && configTargets !== void 0 ? configTargets : [document];
      if (forms) {
          Array.prototype.forEach.call(forms, function (form) {
              targets.push(form.ownerDocument.documentElement);
          });
      }
      return targets;
  }
  /**
   * Remove all submission/focus/change event listeners from page that have been added via a call to `addFormListeners`
   *
   * @param tracker The tracker instance the listener belongs to that will be used to track events
   */
  function removeFormListeners(tracker) {
      var _a;
      var targets = (_a = _targets[tracker.id]) !== null && _a !== void 0 ? _a : [document];
      targets.forEach(function (target) {
          if (_focusListeners[tracker.id])
              target.removeEventListener('focus', _focusListeners[tracker.id], true);
          if (_changeListeners[tracker.id])
              target.removeEventListener('change', _changeListeners[tracker.id], true);
          if (_submitListeners[tracker.id])
              target.removeEventListener('submit', _submitListeners[tracker.id], true);
      });
  }
  /**
   * Check if forms array is a collection of HTML form elements or a filter or undefined
   */
  function isCollectionOfHTMLFormElements(forms) {
      return forms != null && Array.prototype.slice.call(forms).length > 0;
  }
  /**
   * Typeguard for `element` to see if it appears to be the HTMLElement with tagName `type`
   *
   * instanceof checks don't work for cross-document nodes, which this plugin supports
   *
   * @param elem Object to check element type
   * @param type Element type we're checking for
   * @returns If `element` is an element with tagName `type`
   */
  function isElement(elem, type) {
      if (typeof elem === 'object' && elem) {
          if ('tagName' in elem && typeof elem['tagName'] === 'string') {
              return elem.tagName.toUpperCase() === type;
          }
      }
      return false;
  }
  /**
   * Determine if given object is a `TrackedHTMLElement` or not
   *
   * @param element Value to determine
   * @returns If `element` is `TrackedHTMLElement`
   */
  function isTrackableElement(element) {
      return isElement(element, 'INPUT') || isElement(element, 'SELECT') || isElement(element, 'TEXTAREA');
  }
  /**
   * Configures form tracking: which forms and fields will be tracked, and the context to attach
   *
   * @param options User-supplied configuration
   * @returns Final configuration incorporating defaults
   */
  function getConfigurationForOptions(options) {
      if (options) {
          var formFilter = function (_) { return true; };
          var forms = null;
          if (isCollectionOfHTMLFormElements(options.forms)) {
              // options.forms is an explicity allowlist of HTML form elements
              forms = options.forms;
          }
          else {
              // options.forms is null or a filter
              formFilter = getFilterByClass(options.forms);
          }
          return {
              forms: forms,
              formFilter: formFilter,
              fieldFilter: getFilterByName(options.fields),
              fieldTransform: getTransform(options.fields),
          };
      }
      else {
          return {
              forms: null,
              formFilter: function () { return true; },
              fieldFilter: function () { return true; },
              fieldTransform: defaultTransformFn,
          };
      }
  }
  /**
   * Check if the found target element is included in the explicit form allowlist, if provided.
   *
   * @param target A `form` element to check if we're allowed to track.
   * @param allowed An optional list of form elements we want to track against.
   * @returns True if there is no allowlist or the `target` is in the allowlist, false otherwise.
   */
  function explicitlyAllowedForm(target, allowed) {
      if (!allowed)
          return true;
      for (var i = 0; i < allowed.length; i++) {
          if (allowed[i].isSameNode(target))
              return true;
      }
      return false;
  }
  /**
   * Convert a criterion object to a transform function
   *
   * @param criterion
   * @returns Transformation function if provided in `criterion`, or a default identity function
   */
  function getTransform(criterion) {
      if (criterion && typeof criterion.transform === 'function') {
          return criterion.transform;
      }
      return defaultTransformFn;
  }
  /**
   * Get an identifier for a form or `TrackedHTMLElement`
   *
   * @param elt Element to identify
   * @returns Identifier for `elt`
   */
  function getElementIdentifier(elt) {
      var properties = ['name', 'id', 'type', 'nodeName'];
      for (var _i = 0, properties_1 = properties; _i < properties_1.length; _i++) {
          var propName = properties_1[_i];
          if (elt[propName] && typeof elt[propName] === 'string') {
              return elt[propName];
          }
      }
      return null;
  }
  /**
   * Discovers the parent form in which an element is contained
   *
   * @param elt Child control to identify the owning form for
   * @returns The form element this control belongs to or null if not found
   */
  function getParentForm(elt) {
      if (elt && elt.form)
          return elt.form;
      var parent = elt;
      while (parent) {
          if (isElement(parent, 'FORM')) {
              return parent;
          }
          parent = parent.parentNode;
      }
      return parent;
  }
  /**
   * Returns a list of the `TrackedHTMLElement`s inside a form along with their values
   *
   * @param elt Form element to get the control elements for
   * @returns Array of wrapped control elements belonging to the form
   */
  function getInnerFormElements(elt) {
      var innerElements = [];
      Array.prototype.forEach.call(elt.elements, function (child) {
          if (!isTrackableElement(child))
              return;
          var inputType = (child.type || 'text').toLowerCase();
          // submit and image are roughly equivalent
          if (inputType === 'submit' || inputType === 'image') {
              return;
          }
          var elementJson = {
              elementData: {
                  name: getElementIdentifier(child),
                  value: child.value,
                  nodeName: child.nodeName,
              },
              originalElement: child,
          };
          if (isElement(child, 'INPUT')) {
              elementJson.elementData.type = inputType;
              if (inputType === 'password' || ((inputType === 'checkbox' || inputType === 'radio') && !child.checked)) {
                  elementJson.elementData.value = null;
              }
          }
          innerElements.push(elementJson);
      });
      return innerElements;
  }
  /**
   * Create closure function to handle form field change/focus event
   *
   * @param tracker The tracker instance to generate the event with
   * @param config Plugin configuration
   * @param event_type Type of event to generate
   * @param context List of entities or context generators to evaluate with the event
   * @returns A form change/focus handler
   */
  function getFormChangeListener(tracker, config, event_type, context) {
      return function (e) {
          var _a, _b;
          var target = e.composed ? e.composedPath()[0] : e.target;
          // `change` and `submit` are not composed and are thus invisible to us
          // bind late to the forms/field directly on field focus in this case
          if (target !== e.target && e.composed && isTrackableElement(target)) {
              if (target.form) {
                  if (_changeListeners[tracker.id])
                      addEventListener(target.form, 'change', _changeListeners[tracker.id], true);
                  if (_submitListeners[tracker.id])
                      addEventListener(target.form, 'submit', _submitListeners[tracker.id], true);
              }
              else {
                  if (_changeListeners[tracker.id])
                      addEventListener(target, 'change', _changeListeners[tracker.id], true);
              }
          }
          if (isTrackableElement(target) && config.fieldFilter(target)) {
              var value = null;
              var type = null;
              if (isElement(target, 'INPUT')) {
                  type = (target.type || 'text').toLowerCase();
                  value =
                      (type === 'checkbox' && !target.checked) || type === 'password'
                          ? null
                          : config.fieldTransform(target.value, target, target);
              }
              else {
                  value = config.fieldTransform(target.value, target, target);
              }
              var form = getParentForm(target);
              if (!(form && config.formFilter(form) && explicitlyAllowedForm(form, config.forms)))
                  return;
              if (event_type === 'change_form' || (type !== 'checkbox' && type !== 'radio')) {
                  tracker.core.track(buildFormFocusOrChange({
                      schema: event_type,
                      formId: (_a = getElementIdentifier(form !== null && form !== void 0 ? form : {})) !== null && _a !== void 0 ? _a : '',
                      elementId: (_b = getElementIdentifier(target)) !== null && _b !== void 0 ? _b : '',
                      nodeName: target.nodeName,
                      type: type,
                      elementClasses: getCssClasses(target),
                      value: value !== null && value !== void 0 ? value : null,
                  }), resolveDynamicContext(context, target, type, value));
              }
          }
      };
  }
  /**
   * Create closure function to handle form submission event
   *
   * @param tracker The tracker instance to generate the event with
   * @param config Plugin configuration
   * @param context List of entities or context generators to evaluate with the event
   * @returns A form submit handler
   */
  function getFormSubmissionListener(tracker, config, context) {
      return function (_a) {
          var _b;
          var target = _a.target;
          if (isElement(target, 'FORM') && config.formFilter(target) && explicitlyAllowedForm(target, config.forms)) {
              var elementsData_1 = [];
              getInnerFormElements(target).forEach(function (_a) {
                  var elementData = _a.elementData, originalElement = _a.originalElement;
                  if (config.fieldFilter(originalElement) && originalElement.type.toLowerCase() !== 'password') {
                      elementData.value = config.fieldTransform(elementData.value, elementData, originalElement);
                      elementsData_1.push(elementData);
                  }
              });
              tracker.core.track(buildFormSubmission({
                  formId: (_b = getElementIdentifier(target)) !== null && _b !== void 0 ? _b : '',
                  formClasses: getCssClasses(target),
                  elements: elementsData_1,
              }), resolveDynamicContext(context, target, elementsData_1));
              flushPendingCookies();
          }
      };
  }

  var _trackers = {};
  /**
   * A plugin which enables automatic form focus, change, and submit tracking
   */
  function FormTrackingPlugin() {
      return {
          activateBrowserPlugin: function (tracker) {
              _trackers[tracker.id] = tracker;
          },
      };
  }
  /**
   * Enables automatic form tracking.
   *
   * An event will be fired when a form field is changed or a form submitted.
   * This can be called multiple times: previous listeners will be removed and replaced with any updated configuration.
   *
   * @param configuration - The form tracking configuration
   * @param trackers - The tracker identifiers which the events will be sent to
   */
  function enableFormTracking(configuration, trackers) {
      if (configuration === void 0) { configuration = {}; }
      if (trackers === void 0) { trackers = Object.keys(_trackers); }
      trackers.forEach(function (t) {
          if (_trackers[t]) {
              removeFormListeners(_trackers[t]);
              addFormListeners(_trackers[t], configuration);
          }
      });
  }
  /**
   * Disables automatic form tracking.
   *
   * All page-level listeners for the given trackers will be removed.
   *
   * @param trackers - The tracker identifiers which the events will be sent to
   */
  function disableFormTracking(trackers) {
      if (trackers === void 0) { trackers = Object.keys(_trackers); }
      trackers.forEach(function (t) {
          if (_trackers[t]) {
              removeFormListeners(_trackers[t]);
          }
      });
  }

  exports.FormTrackingPlugin = FormTrackingPlugin;
  exports.disableFormTracking = disableFormTracking;
  exports.enableFormTracking = enableFormTracking;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
//# sourceMappingURL=index.umd.js.map
